# This file is a part of the `nequip` package. Please see LICENSE and README at the root for information on using it.
import torch

# singular source of global dtype that dictates the dtype of the data, which is always float64
_GLOBAL_DTYPE = torch.float64
